import os
import random
import natsort
import datetime
import dateutil


from PIL import Image
from io import BytesIO
from operator import itemgetter
from typing import Any, Dict, List
from chocolate_app import get_language_file
from chocolate_app.routes.api.auth import token_required
from flask import Blueprint, request, Response, abort, send_file
from chocolate_app.utils.utils import generate_response, Codes, translate

from chocolate_app.tables import (
    TVChannels,
    TVPrograms,
    Users,
    Series,
    Movies,
    Actors,
    Albums,
    Artists,
    Episodes,
    Seasons,
    Tracks,
    Games,
    Books,
    OthersVideos,
    MediaPlayed,
)

medias_bp = Blueprint("medias", __name__, url_prefix="/medias")

LANGUAGE_FILE = get_language_file()


def extract_sibling_episodes(episode, main_season):
    previous = None
    next = None
    episodes = natsort.natsorted(
        Episodes.query.filter_by(season_id=main_season.tmdb_id).all(),
        key=lambda x: x.release_date,
    )
    for i, ep in enumerate(episodes):
        if ep.id == episode.id:
            if i != 0:
                previous = episodes[i - 1]
            else:
                previous_season = Seasons.query.filter_by(
                    tmdb_id=main_season.serie_id, number=main_season.number - 1
                ).first()
                if previous_season:
                    previous_episodes = natsort.natsorted(
                        Episodes.query.filter_by(
                            season_id=previous_season.tmdb_id
                        ).all(),
                        key=lambda x: x.release_date,
                    )
                    previous = previous_episodes[-1]
            if i != len(episodes) - 1:
                next = episodes[i + 1]
            else:
                next_season = Seasons.query.filter_by(
                    tmdb_id=main_season.serie_id, number=main_season.number + 1
                ).first()
                if next_season:
                    next_episodes = natsort.natsorted(
                        Episodes.query.filter_by(season_id=next_season.tmdb_id).all(),
                        key=lambda x: x.release_date,
                    )
                    next = next_episodes[0]
            break

    return previous, next


def episode_to_media(
    user_id, episode_id, have_serie_repr=True
) -> Dict[str, Any] | None:
    episode = Episodes.query.filter_by(id=episode_id).first()
    if not episode:
        return None
    main_season = Seasons.query.filter_by(tmdb_id=episode.season_id).first()
    if not main_season:
        return None
    serie = Series.query.filter_by(tmdb_id=main_season.serie_id).first()
    if not serie:
        return None

    seasons = natsort.natsorted(
        Seasons.query.filter_by(serie_id=serie.tmdb_id).all(), key=lambda x: x.release
    )

    # TODO: Don't select the first one, select the last you watched

    banner_b64 = episode.cover_b64
    logo_b64 = serie.logo_b64
    cover_b64 = serie.cover_b64

    actors = serie.cast.split(",")
    actor_list = []

    for actor in actors:
        actor = Actors.query.filter_by(id=actor).first()
        if actor:
            actor_list.append({"name": actor.name, "type": "actor"})

    serie_representation = []

    if have_serie_repr:
        for season in seasons:
            episodes = natsort.natsorted(
                Episodes.query.filter_by(season_id=season.tmdb_id).all(),
                key=lambda x: x.release_date,
            )
            eps = []

            for ep in episodes:
                eps.append(episode_to_media(user_id, ep.id, False))

            serie_representation.append(
                {
                    "season_id": season.id,
                    "season_number": season.number,
                    "episodes": eps,
                }
            )

    previous, next = extract_sibling_episodes(episode, main_season)

    last_duration = 0
    try:
        media_played = MediaPlayed.query.filter_by(
            user_id=user_id, media_id=episode_id, media_type="show"
        ).first()
        if media_played:
            last_duration = media_played.duration
    except Exception:
        pass

    media = {
        "id": episode_id,
        "title": episode.title,
        "serie_title": serie.title,
        "alternatives_titles": [serie.title],
        "banner_id": serie.id,
        "description": serie.description,
        "have_logo": True if serie.logo else False,
        "type": "show",
        "note": serie.note,
        "genres": [translate(genre) for genre in serie.genre.split(",")],
        "duration": len(Episodes.query.filter_by(season_id=main_season.tmdb_id).all()),
        "release_date": datetime.datetime.strptime(serie.date, "%Y-%m-%d"),
        "file_date": datetime.datetime.fromtimestamp(os.path.getmtime(episode.slug)),
        "serie_representation": serie_representation,
        "last_duration": last_duration,
        "number": episode.number,
        "images": {
            "banner": banner_b64,
            "logo": logo_b64,
            "cover": cover_b64,
        },
        "peoples": actor_list,
        "_previous": previous.id if previous else None,
        "_next": next.id if next else None,
    }

    return media


def serie_to_media(user_id, serie_id) -> Dict[str, Any] | None:
    serie = Series.query.filter_by(id=serie_id).first()
    if not serie:
        return None

    serie_media_played = MediaPlayed.query.filter_by(
        user_id=user_id, serie_id=serie_id, media_type="show"
    ).all()

    episode_exists = Episodes.query.filter_by(serie_id=serie.tmdb_id).first()

    if not serie_media_played or not episode_exists:
        if not episode_exists:
            for media in serie_media_played:
                MediaPlayed.query.filter_by(id=media.id).delete()
        return episode_to_media(
            user_id,
            natsort.natsorted(
                Episodes.query.filter_by(serie_id=serie.tmdb_id).all(),
                key=lambda x: x.release_date,
            )[0].id,
            False,
        )

    serie_media_played = [media.__dict__ for media in serie_media_played]

    serie_media_played = natsort.natsorted(
        serie_media_played, key=itemgetter(*["datetime"]), reverse=True
    )

    episode_id = serie_media_played[0]["media_id"]
    episode_exists = Episodes.query.filter_by(id=episode_id).first()

    if not episode_exists:
        for media in serie_media_played:
            MediaPlayed.query.filter_by(id=media["id"]).delete()
        return episode_to_media(
            user_id,
            natsort.natsorted(
                Episodes.query.filter_by(serie_id=serie.tmdb_id).all(),
                key=lambda x: x.release_date,
            )[0].id,
            False,
        )

    return episode_to_media(user_id, episode_id, False)


def movie_to_media(user_id, movie_id) -> Dict[str, Any] | None:
    movie = Movies.query.filter_by(id=movie_id).first()
    if not movie:
        return None
    actors = movie.cast.split(",")
    actor_list = []

    for actor in actors:
        actor = Actors.query.filter_by(id=actor).first()
        if actor:
            actor_list.append({"name": actor.name, "type": "actor"})

    banner_b64 = movie.banner_b64
    logo_b64 = movie.logo_b64
    cover_b64 = movie.cover_b64

    last_duration = 0
    try:
        media_played = MediaPlayed.query.filter_by(
            user_id=user_id, media_id=movie_id, media_type="movie"
        ).first()
        if media_played:
            last_duration = media_played.duration
    except Exception:
        pass

    date = movie.date

    if date == "":
        date = datetime.datetime.now()
    else:
        date = datetime.datetime.strptime(date, "%Y-%m-%d")

    media = {
        "id": movie.id,
        "title": movie.title,
        "alternatives_titles": movie.alternative_title.split(","),
        "banner_id": movie.id,
        "description": movie.description,
        "have_logo": True if movie.logo else False,
        "type": "movie",
        "note": movie.note,
        "genres": [translate(genre) for genre in movie.genre.split(",")],
        "duration": movie.duration,
        "release_date": date,
        "file_date": datetime.datetime.fromtimestamp(movie.file_date),
        "last_duration": last_duration,
        "images": {
            "banner": banner_b64,
            "logo": logo_b64,
            "cover": cover_b64,
        },
        "peoples": actor_list,
    }
    return media


def other_to_media(user_id, other_id) -> Dict[str, Any] | None:
    other = OthersVideos.query.filter_by(id=other_id).first()
    if not other:
        return None
    banner_b64 = other.banner_b64
    logo_b64 = banner_b64
    cover_b64 = banner_b64

    last_duration = 0
    media_played = MediaPlayed.query.filter_by(
        user_id=user_id, media_id=other_id, media_type="other"
    ).first()
    if media_played:
        last_duration = media_played.duration

    media = {
        "id": other.video_hash,
        "title": other.title,
        "alternatives_titles": [other.title],
        "banner_id": other.video_hash,
        "description": "",
        "have_logo": False,
        "type": "other",
        "note": 0,
        "genres": [],
        "duration": other.duration,
        "release_date": "N/A",
        "file_date": datetime.datetime.fromtimestamp(os.path.getmtime(other.slug)),
        "last_duration": last_duration,
        "images": {
            "banner": banner_b64,
            "logo": logo_b64,
            "cover": cover_b64,
        },
        "peoples": [],
    }
    return media


def album_to_media(user_id, album_id) -> Dict[str, Any] | None:
    album = Albums.query.filter_by(id=album_id).first()
    if not album:
        return None
    artist = Artists.query.filter_by(id=album.artist_id).first()
    if not artist:
        return None
    track = Tracks.query.filter_by(album_id=album_id).first()
    if not track:
        return None

    cover_b64 = album.cover_b64
    banner_b64 = cover_b64

    media = {
        "id": album.id,
        "title": album.title,
        "alternatives_titles": [],
        "banner_id": album.id,
        "description": "",
        "have_logo": False,
        "type": "album",
        "note": 0,
        "genres": [translate(10402)],
        "duration": len(Tracks.query.filter_by(album_id=album_id).all()),
        "release_date": datetime.datetime.strptime(album.release_date, "%Y-%m-%d"),
        "file_date": datetime.datetime.fromtimestamp(track.file_date),
        "images": {
            "banner": banner_b64,
            "logo": None,
            "cover": cover_b64,
        },
        "peoples": [{"name": artist.name, "type": "artist"}],
    }
    return media


def get_current_program(channel_id) -> Any | None:
    now = datetime.datetime.now(datetime.timezone.utc)
    programs = TVPrograms.query.filter_by(channel_id=channel_id).all()
    for program in programs:
        start = dateutil.parser.parse(program.start_time)
        end = dateutil.parser.parse(program.end_time)
        program = program.__dict__
        del program["_sa_instance_state"]
        if start <= now <= end:
            return program
    return None


def get_sibling_channels(channel_id) -> Any:
    channel = TVChannels.query.filter_by(id=channel_id).first()
    if not channel:
        return None, None

    # get the first channel with an id lower than the current channel
    previous = (
        TVChannels.query.filter(TVChannels.id < channel_id)
        .order_by(TVChannels.id.desc())
        .first()
    )

    # get the first channel with an id higher than the current channel
    next = (
        TVChannels.query.filter(TVChannels.id > channel_id)
        .order_by(TVChannels.id.asc())
        .first()
    )

    return previous, next


def tv_to_media(channel_id) -> Dict[str, Any] | None:
    channel = TVChannels.query.filter_by(id=channel_id).first()

    if not channel:
        return None

    previous, next = get_sibling_channels(channel_id)

    media = {
        "id": f"{channel.id}",
        "_source": channel.slug,
        "_epg": get_current_program(channel_id),
        "title": channel.name,
        "alternatives_titles": [channel.name],
        "banner_id": channel.id,
        "description": "",
        "have_logo": True,
        "type": "live-tv",
        "note": 0,
        "genres": [],
        "duration": 0,
        "release_date": "N/A",
        "file_date": "N/A",
        "images": {
            "banner": channel.logo,
            "logo": channel.logo,
            "cover": channel.logo,
        },
        "peoples": [],
        "_previous": previous.id if previous else None,
        "_next": next.id if next else None,
    }

    return media


def check_usability(media_list: List[Dict[str, Any]]) -> None:
    # for all media, check if the id is in the database
    for media in media_list:
        if media["media_type"] == "movie":
            movie = Movies.query.filter_by(id=media["id"]).first()
            if not movie:
                media_list.remove(media)
        elif media["media_type"] == "show":
            serie = Series.query.filter_by(id=media["id"]).first()
            if not serie:
                media_list.remove(media)
        elif media["media_type"] == "album":
            album = Albums.query.filter_by(id=media["id"]).first()
            if not album:
                media_list.remove(media)
        elif media["media_type"] == "other":
            other = OthersVideos.query.filter_by(video_hash=media["id"]).first()
            if not other:
                media_list.remove(media)


def get_continue_watching(user: Users, media_type) -> List[Dict[str, Any]]:
    user_id = user.id
    if media_type == "all":
        continue_watching = MediaPlayed.query.filter_by(user_id=user_id).all()
    else:
        continue_watching = MediaPlayed.query.filter_by(
            user_id=user_id, media_type=media_type
        ).all()
    continue_watching_list = [media.__dict__ for media in continue_watching]

    check_usability(continue_watching_list)

    for watching in continue_watching_list:
        del watching["_sa_instance_state"]

    serie_ids_in_continue = []

    # sort by date, and then by time
    continue_watching_list = natsort.natsorted(
        continue_watching_list, key=itemgetter(*["datetime"]), reverse=True
    )
    continue_watching_list_to_return = []

    type_to_function = {
        "movie": movie_to_media,
        "show": episode_to_media,
        "album": album_to_media,
        "other": other_to_media,
    }

    for watching in continue_watching_list:
        if (
            watching["media_type"] == "show"
            and watching["serie_id"] in serie_ids_in_continue
        ):
            continue
        if watching["media_type"] == "show":
            serie_ids_in_continue.append(watching["serie_id"])
        media_type = watching["media_type"]
        if media_type == "show":
            continue_watching_list_to_return.append(
                episode_to_media(user_id, watching["media_id"], False)
            )
        else:
            continue_watching_list_to_return.append(
                type_to_function[watching["media_type"]](user_id, watching["media_id"])
            )

    return continue_watching_list_to_return


def get_all_medias_without_albums(user_id) -> List[Dict[str, Any]]:
    all_movies = [movie_to_media(user_id, movie.id) for movie in Movies.query.all()]

    all_series_db = Series.query.all()
    all_series = []

    for serie in all_series_db:
        try:
            media = serie_to_media(user_id, serie.id)
        except Exception:
            continue
        if media:
            all_series.append(media)

    all_medias = all_movies + all_series

    return all_medias  # type: ignore


def get_all_medias(user_id) -> Dict[str, List[Dict[str, Any]]]:
    all_medias_without_albums = get_all_medias_without_albums(user_id)
    all_albums = [album_to_media(user_id, album.id) for album in Albums.query.all()]
    all_albums = [album for album in all_albums if album is not None]
    return all_medias_without_albums + all_albums  # type: ignore


def get_medias_for_key(key: str, all_medias: List[Dict[str, Any]]) -> List[str]:
    medias = [media for media in all_medias if media is not None]
    medias = natsort.natsorted(medias, key=itemgetter(*[key]), reverse=True)
    return medias[:15]  # type: ignore


def get_latest_scanned_medias(all_medias: List[Dict[str, Any]]) -> List[str]:
    return get_medias_for_key("file_date", all_medias)


def get_latest_medias(all_medias: List[Dict[str, Any]]) -> List[str]:
    return get_medias_for_key("release_date", all_medias)


def get_top_rated_medias(all_medias: List[Dict[str, Any]]) -> List[str]:
    return get_medias_for_key("note", all_medias)


def get_best_of_year(all_medias: List[Dict[str, Any]]) -> List[str]:
    current_year = datetime.datetime.now().year

    medias = [media for media in all_medias if media is not None]
    medias = [
        media for media in medias if media["release_date"].year == current_year  # type: ignore
    ]

    medias = natsort.natsorted(medias, key=itemgetter(*["note"]), reverse=True)

    return medias[:15]  # type: ignore


def get_media_for_genre(genre: int, all_medias: list[Dict[str, Any]]) -> List[str]:
    medias = [media for media in all_medias if media is not None]
    medias = [media for media in medias if int(genre) in media["genres"]]  # type: ignore
    medias = natsort.natsorted(medias, key=itemgetter(*["release_date"]))
    return medias[:15]  # type: ignore


@medias_bp.route("home", methods=["GET"])
@token_required
def get_home_medias(current_user) -> Response:
    data = {
        "main_media": None,
        "continue_watching": [],
        "latest": [],
        "recently_added": [],
        "top_rated": [],
        "best_of_year": [],
        "family": [],
        "comedy": [],
        "animated": [],
        "action": [],
        "thriller": [],
        "horror": [],
        "drama": [],
        "western": [],
    }

    all_medias: Any = get_all_medias_without_albums(current_user.id)
    albums = [album_to_media(current_user.id, album.id) for album in Albums.query.all()]
    albums = [album for album in albums if album is not None]

    continue_watching = get_continue_watching(current_user, "all")
    data["continue_watching"] = continue_watching

    latest = get_latest_medias(all_medias + albums)  # type: ignore
    data["latest"] = latest

    recently_added = get_latest_scanned_medias(all_medias + albums)  # type: ignore
    data["recently_added"] = recently_added

    top_rated = get_top_rated_medias(all_medias)
    data["top_rated"] = top_rated

    best_of_year = get_best_of_year(all_medias)
    data["best_of_year"] = best_of_year

    family = get_media_for_genre(10751, all_medias)
    data["family"] = family

    comedy = get_media_for_genre(35, all_medias)
    data["comedy"] = comedy

    animated = get_media_for_genre(16, all_medias)
    data["animated"] = animated

    action = get_media_for_genre(28, all_medias)
    data["action"] = action

    thriller = get_media_for_genre(53, all_medias)
    data["thriller"] = thriller

    horror = get_media_for_genre(27, all_medias)
    data["horror"] = horror

    drama = get_media_for_genre(18, all_medias)
    data["drama"] = drama

    western = get_media_for_genre(37, all_medias)
    data["western"] = western

    keys = list(data.keys())
    keys.pop(0)

    for key in keys:
        medias = data[key]
        for media in medias:
            if media in all_medias:
                all_medias.remove(media)

    if len(all_medias) > 0:
        data["main_media"] = random.choice(all_medias)  # type: ignore
    else:
        all_medias = get_all_medias_without_albums(current_user.id)
        if len(all_medias) == 0:
            return generate_response(Codes.MEDIA_NOT_FOUND, True)
        data["main_media"] = all_medias[0]  # type: ignore

    return generate_response(Codes.SUCCESS, False, data)


@medias_bp.route("movies", methods=["GET"])
@token_required
def get_movies_media(current_user) -> Response:
    data = {
        "main_media": None,
        "continue_watching": [],
        "latest": [],
        "recently_added": [],
        "top_rated": [],
        "best_of_year": [],
        "family": [],
        "comedy": [],
        "animated": [],
        "action": [],
        "thriller": [],
        "horror": [],
        "drama": [],
        "western": [],
    }

    all_medias = [
        movie_to_media(current_user.id, movie.id) for movie in Movies.query.all()
    ]
    all_medias: Any = [media for media in all_medias if media is not None]

    continue_watching = get_continue_watching(current_user, "movie")
    data["continue_watching"] = continue_watching

    latest = get_latest_medias(all_medias)
    data["latest"] = latest

    recently_added = get_latest_scanned_medias(all_medias)
    data["recently_added"] = recently_added

    top_rated = get_top_rated_medias(all_medias)
    data["top_rated"] = top_rated

    best_of_year = get_best_of_year(all_medias)
    data["best_of_year"] = best_of_year

    family = get_media_for_genre(10751, all_medias)
    data["family"] = family

    comedy = get_media_for_genre(35, all_medias)
    data["comedy"] = comedy

    animated = get_media_for_genre(16, all_medias)
    data["animated"] = animated

    action = get_media_for_genre(28, all_medias)
    data["action"] = action

    thriller = get_media_for_genre(53, all_medias)
    data["thriller"] = thriller

    horror = get_media_for_genre(27, all_medias)
    data["horror"] = horror

    drama = get_media_for_genre(18, all_medias)
    data["drama"] = drama

    western = get_media_for_genre(37, all_medias)
    data["western"] = western

    keys = list(data.keys())
    keys.pop(0)

    for key in keys:
        medias = data[key]
        for media in medias:
            if media in all_medias:
                all_medias.remove(media)

    if len(all_medias) > 0:
        data["main_media"] = random.choice(all_medias)
    else:
        medias = get_all_medias_without_albums(current_user.id)
        if len(medias) > 0:
            data["main_media"] = medias[0]
        else:
            data["main_media"] = None

    return generate_response(Codes.SUCCESS, False, data)


@medias_bp.route("shows", methods=["GET"])
@token_required
def get_shows_media(current_user) -> Response:
    data = {
        "main_media": None,
        "continue_watching": [],
        "latest": [],
        "recently_added": [],
        "top_rated": [],
        "best_of_year": [],
        "family": [],
        "comedy": [],
        "animated": [],
        "action": [],
        "thriller": [],
        "horror": [],
        "drama": [],
        "western": [],
    }

    all_medias = [
        serie_to_media(current_user.id, serie.id) for serie in Series.query.all()
    ]
    all_medias = [media for media in all_medias if media is not None]

    continue_watching = get_continue_watching(current_user, "show")
    data["continue_watching"] = continue_watching

    latest = get_latest_medias(all_medias)
    data["latest"] = latest

    recently_added = get_latest_scanned_medias(all_medias)
    data["recently_added"] = recently_added

    top_rated = get_top_rated_medias(all_medias)
    data["top_rated"] = top_rated

    best_of_year = get_best_of_year(all_medias)
    data["best_of_year"] = best_of_year

    family = get_media_for_genre(10751, all_medias)
    data["family"] = family

    comedy = get_media_for_genre(35, all_medias)
    data["comedy"] = comedy

    animated = get_media_for_genre(16, all_medias)
    data["animated"] = animated

    action = get_media_for_genre(28, all_medias)
    data["action"] = action

    thriller = get_media_for_genre(53, all_medias)
    data["thriller"] = thriller

    horror = get_media_for_genre(27, all_medias)
    data["horror"] = horror

    drama = get_media_for_genre(18, all_medias)
    data["drama"] = drama

    western = get_media_for_genre(37, all_medias)
    data["western"] = western

    keys = list(data.keys())
    keys.pop(0)

    for key in keys:
        medias = data[key]
        for media in medias:
            if media in all_medias:
                all_medias.remove(media)

    if len(all_medias) > 0:
        data["main_media"] = random.choice(all_medias)
    else:
        medias = get_all_medias_without_albums(current_user.id)
        if len(medias) > 0:
            data["main_media"] = medias[0]
        else:
            data["main_media"] = None

    return generate_response(Codes.SUCCESS, False, data)


@medias_bp.route("tv", methods=["GET"])
@token_required
def get_tv_media(current_user) -> Response:
    data = {"medias": []}

    data["medias"] = [tv_to_media(channel.id) for channel in TVChannels.query.all()]

    # on trie par ordre alphabétique
    data["medias"] = natsort.natsorted(data["medias"], key=lambda x: x["title"])

    return generate_response(Codes.SUCCESS, False, data)


@medias_bp.route("/images/<image_type>/<media_type>/<media_id>", methods=["GET"])
def get_image(image_type: str, media_type: str, media_id: int) -> Response:
    image = None
    if image_type == "banner":
        if media_type == "movie":
            movie = Movies.query.filter_by(id=media_id).first()
            if not movie:
                abort(404)
            image = movie.banner
        elif media_type == "show":
            serie = Series.query.filter_by(id=media_id).first()
            if not serie:
                abort(404)
            image = serie.banner
        elif media_type == "episode":
            ep = Episodes.query.filter_by(episode_id=media_id).first()
            if not ep:
                abort(404)
            serie = Series.query.filter_by(tmdb_id=ep.serie_id).first()
            if not serie:
                abort(404)
            image = serie.banner
        elif media_type == "album":
            album = Albums.query.filter_by(id=media_id).first()
            if not album:
                abort(404)
            image = album.cover
        elif media_type == "game":
            game = Games.query.filter_by(id=media_id).first()
            if not game:
                abort(404)
        elif media_type == "books":
            book = Books.query.filter_by(id=media_id).first()
            if not book:
                abort(404)
            image = book.cover
        else:
            abort(404)
    elif image_type == "cover":
        if media_type == "movie":
            movie = Movies.query.filter_by(id=media_id).first()
            if not movie:
                abort(404)
            image = movie.cover
        elif media_type == "show":
            serie = Series.query.filter_by(id=media_id).first()
            if not serie:
                abort(404)
            image = serie.cover
        elif media_type == "episode":
            ep = Episodes.query.filter_by(episode_id=media_id).first()
            if not ep:
                abort(404)
            serie = Series.query.filter_by(tmdb_id=ep.serie_id).first()
            if not serie:
                abort(404)
            image = serie.cover
        elif media_type == "album":
            album = Albums.query.filter_by(id=media_id).first()
            if not album:
                abort(404)
            image = album.cover
        elif media_type == "game":
            game = Games.query.filter_by(id=media_id).first()
            if not game:
                abort(404)
            image = game.cover
        elif media_type == "books":
            book = Books.query.filter_by(id=media_id).first()
            if not book:
                abort(404)
            image = book.cover
        else:
            abort(404)
    elif image_type == "logo":
        if media_type == "show" or media_type == "episode":
            serie = Series.query.filter_by(id=media_id).first()
            if not serie:
                abort(404)
            image = serie.logo
        elif media_type == "movie":
            movie = Movies.query.filter_by(id=media_id).first()
            if not movie:
                abort(404)
            image = movie.logo
        else:
            abort(404)

    if not image:
        abort(404)

    extension = os.path.splitext(image)[1]
    mime_type = {
        ".jpg": "image/jpeg",
        ".jpeg": "image/jpeg",
        ".png": "image/png",
        ".webp": "image/webp",
    }

    width = request.args.get("width")
    if width:
        image = Image.open(image)
        wpercent = int(width) / float(image.size[0])
        hsize = int((float(image.size[1]) * float(wpercent)))
        image = image.resize((int(width), hsize), Image.ANTIALIAS)  # type: ignore
        img_io = BytesIO()
        image.save(img_io, "WEBP")
        img_io.seek(0)
        return send_file(img_io, mimetype=mime_type.get(extension, "image/webp"))

    height = request.args.get("height")
    if height:
        image = Image.open(image)
        hpercent = int(height) / float(image.size[1])
        wsize = int((float(image.size[0]) * float(hpercent)))
        image = image.resize((wsize, int(height)), Image.ANTIALIAS)  # type: ignore
        img_io = BytesIO()
        image.save(img_io, "WEBP")
        img_io.seek(0)
        return send_file(img_io, mimetype=mime_type.get(extension, "image/webp"))

    try:
        return send_file(image, mimetype=mime_type.get(extension, "image/jpeg"))
    except FileNotFoundError:
        return send_file(
            f"static/img/broken{ 'Banner' if image_type == 'banner' else '' }.webp",
            mimetype="image/webp",
        )


def genre_id_to_name(genre_id: int) -> str:
    return translate(genre_id)


def search_medias(
    medias: List[Dict[str, Any]], search_terms: List[str], user_id
) -> Dict[str, List[Dict[str, Any]]]:
    results = {}
    for media in medias:
        if not media:
            continue
        count = 0.0

        title = media["title"].lower()
        alternative_titles = media["alternatives_titles"]
        description = media["description"].lower().split(" ")
        genres = media["genres"]
        people = media["peoples"]

        for term in search_terms:
            term = term.lower()
            if term in title:
                count += 2
            for alt_title in alternative_titles:
                if term in alt_title.lower():
                    count += 1
            for word in description:
                if term == word.lower():
                    count += 0.1
            for genre in genres:
                if term == genre.lower():
                    count += 0.5
            for person in people:
                if term == person["name"].lower():
                    count += 1
        if count > 0:
            results[f"{media['id']} - {media['title']}"] = count

    temp_results = sorted(results.items(), key=lambda x: x[1], reverse=True)

    results = []
    for result in temp_results:
        for media in medias:
            if f"{media['id']} - {media['title']}" == result[0]:
                results.append(media)
                break

    return results  # type: ignore


def search_movies(user_id, search_terms: List[str]) -> Dict[str, List[Dict[str, Any]]]:
    movies = [movie_to_media(user_id, movie.id) for movie in Movies.query.all()]

    results = search_medias(movies, search_terms, user_id)  # type: ignore

    return results


def search_series(user_id, search_terms: List[str]) -> Dict[str, List[Dict[str, Any]]]:
    series = [serie_to_media(user_id, serie.id) for serie in Series.query.all()]

    results = search_medias(series, search_terms, user_id)  # type: ignore

    return results


def search_musics(user_id, search_terms: List[str]) -> Dict[str, List[Dict[str, Any]]]:
    albums = [album_to_media(user_id, album.id) for album in Albums.query.all()]

    results = search_medias(albums, search_terms, user_id)  # type: ignore

    return results


@medias_bp.route("/search/<type>", methods=["GET"])
@token_required
def search_medias_route(current_user, type: str) -> Response:
    if type not in ["home", "movies", "shows", "musics"]:
        abort(404)

    search = request.args.get("search")

    if not search:
        abort(400, "Missing search")
    search_results = []
    search_terms = search.split(" ")

    if type == "home":
        search_results = search_medias(get_all_medias(current_user.id), search_terms, current_user.id)  # type: ignore
    elif type == "movies":
        search_results = search_movies(current_user.id, search_terms)
    elif type == "shows":
        search_results = search_series(current_user.id, search_terms)
    elif type == "musics":
        search_results = search_musics(current_user.id, search_terms)

    if len(search_results) == 0:
        return generate_response(Codes.NO_RETURN_DATA, False)

    if len(search_results) > 1:
        data = {
            "main_media": search_results[0],  # type: ignore
            "medias": search_results[1:],  # type: ignore
        }
    elif len(search_results) == 1:
        data = {"main_media": search_results[0], "medias": []}  # type: ignore
    else:
        data = {"main_media": None, "medias": []}

    return generate_response(Codes.SUCCESS, False, data)


@medias_bp.route("/media/<media_type>/<media_id>", methods=["GET"])
@token_required
def get_media(current_user, media_type: str, media_id: int) -> Response:
    media = None
    if media_type == "movie":
        media = movie_to_media(current_user.id, media_id)
    elif media_type == "show":
        media = episode_to_media(current_user.id, media_id)
    elif media_type == "album":
        media = album_to_media(current_user.id, media_id)
    elif media_type == "other":
        media = other_to_media(current_user.id, media_id)
    elif media_type == "live-tv":
        media = tv_to_media(media_id)
    else:
        abort(404)

    if not media:
        abort(404)

    return generate_response(Codes.SUCCESS, False, media)



from flask import Blueprint, request

from chocolate_app import DB, scans
from chocolate_app.tables import Libraries, Users
from chocolate_app.routes.api.auth import token_required
from chocolate_app.utils.utils import generate_response, Codes

lib_bp = Blueprint("libraries", __name__, url_prefix="/libraries")


@lib_bp.route("/refresh/<lib_id>", methods=["GET", "POST"])
@token_required
def refresh(current_user, lib_id):
    if not current_user.account_type.lower() == "admin":
        return generate_response(Codes.NOT_ADMIN, True)

    lib = Libraries.query.filter_by(id=lib_id).first()

    if not lib:
        return generate_response(Codes.LIBRARY_NOT_FOUND, True)

    library = lib.__dict__

    type_to_call = {
        "series": scans.getSeries,
        "consoles": scans.getGames,
        "others": scans.getOthersVideos,
        "books": scans.getBooks,
        "musics": scans.getMusics,
    }

    scanner = None
    if library["type"] == "movies":
        scanner = scans.MovieScanner()
        scanner.set_library_name(library["name"])
        scanner.scan()
    elif library["type"] == "tv":
        scanner = scans.LiveTVScanner()
        scanner.set_library_name(library["name"])
        scanner.scan()
    else:
        scanner = type_to_call[library["type"]]
        scanner(library["name"])

    return generate_response(Codes.SUCCESS, False, {"message": "Library refreshed"})



import os
import itertools
import operator
import numpy as np
import pickle
import faiss
import pathlib
import datetime
from natsort import natsorted, ns

from chocolate_app import ARTEFACTS_PATH

# internal imports
from . import featurevectors
from . import video_functions
from . import evaluation


def max_two_values(d):
    """
    a) create a list of the dict's keys and values;
    b) return the two keys with the max values
    """
    v = list(d.values())
    k = list(d.keys())
    result1 = k[v.index(max(v))]
    del d[result1]

    v = list(d.values())
    k = list(d.keys())
    result2 = k[v.index(max(v))]
    return [result1, result2]


def fill_gaps(sequence, lookahead):
    """
    Given a list consisting of 0's and 1's , fills up the gaps between 1's
    if the gap is smaller than the lookahead.

    Example:
        input: [0,0,1,0,0,0,0,1,0,0] with lookahead=6
       output: [0,0,1,1,1,1,1,1,0,0]
    """

    i = 0
    change_needed = False
    look_left = 0
    while i < len(sequence):
        look_left -= 1
        if change_needed and look_left < 1:
            change_needed = False
        if sequence[i]:
            if change_needed:
                for k in to_change:
                    sequence[k] = True
            else:
                change_needed = True
            look_left = lookahead
            to_change = []
        else:
            if change_needed:
                to_change.append(i)
        i += 1
    return sequence


def get_two_longest_timestamps(timestamps):
    """
    Returns the two longest time intervals given a list of time intervals

    Example:
        input: [(0,10) , (0,5) , (20,21)]
        returns: [(0,10), (0,5)]
    """
    # if size is smaller or equal to 2, return immediately
    if len(timestamps) <= 2:
        return timestamps

    d = {}
    for start, end in timestamps:
        d[(start, end)] = end - start

    return max_two_values(d)


def to_time_string(seconds):
    """
    Given seconds in integer format, returns a string in the format hh:mm:ss (example: 01:30:45)
    """
    return str(datetime.timedelta(seconds=seconds))


def query_episodes_with_faiss(videos, vectors_dir):
    """
    Given a vector with the video file names and the directory
    where the corresponding vectors files reside. This function will
    query each set of episode feature vectors on all of the other feature vectors.
    It will return the distances to the best match found on each frame.

    returns:
        A list with tuples consisting of:
        (video_file_name, [list with all distances best match on each frame])
    """
    vector_files = []
    for video in videos:
        file_name, extension = os.path.splitext(video)
        file_name = os.path.basename(file_name)
        file_name = f"{file_name}{extension}.p"
        vector_files.append(file_name)

    vectors = []

    # the lengths of each vector, will be used to query each episode
    lengths = []

    # concatenate all the vectors into a single list multidimensional array
    for f in vector_files:
        episode_vectors = np.array(
            pickle.load(open(f"{ARTEFACTS_PATH}/{f}", "rb")), np.float32
        )
        lengths.append(episode_vectors.shape[0])
        vectors.append(episode_vectors)

    vectors = np.vstack(vectors)

    results = []
    for i, length in enumerate(lengths):
        i += 1
        s = sum(lengths[: i - 1])
        e = sum(lengths[:i])

        # query consists of one episode
        query = vectors[s:e]
        # rest of the feature vectors
        rest = np.append(vectors[:s], vectors[e:], axis=0)

        # build the faiss index, set vector size
        vector_size = query.shape[1]
        index = faiss.IndexFlatL2(vector_size)

        # add vectors of the rest of the episodes to the index
        index.add(rest)

        # we want to see k nearest neighbors
        k = 1
        # search with for matches with query
        scores, indexes = index.search(query, k)

        result = scores[:, 0]
        try:
            results.append((videos[i - 1], result))
        except:
            print(f"Error in query_episodes_with_faiss for idx {i-1}, result: {result}")

    return results


def detect(
    videos,
    feature_vector_function="CH",
    annotations=None,
    artifacts_dir=None,
    framejump=3,
    percentile=10,
    resize_width=144,
    video_start_threshold_percentile=20,
    video_end_threshold_seconds=15,
    min_detection_size_seconds=15,
):
    """
    The main function to call to detect recurring content. Resizes videos, converts to feature vectors
    and returns the locations of recurring content within the videos.

    arguments
    ---------
    videos : list
        Variable that should have the video file locations.
    annotations : str
        Location of the annotations.csv file, if annotations is given then it will evaluate the detections with the annotations.
    feature_vector_function : str
        Which type of feature vectors to use, options: ["CH", "CTM", "CNN"], default is color histograms (CH) because
        of balance between speed and accuracy. This default is defined in init.py.
    artifacts_dir : str
        Directory location where the artifacts should be saved. Default location is the location
        defined with the video_dir parameter.
    framejump : int
        The frame interval to use when sampling frames for the detection, a higher number means that less frames will be
        taken into consideration and will improve the processing time. But will probably cost accuracy.
    percentile : int
        Which percentile of the best matches will be taken into consideration as recurring content. A high percentile will
        means a higher recall, lower precision. A low percentile means a lower recall and higher precision.
    resize_width: int
        Width to which the videos will be resized. A lower number means higher processing speed but less accuracy and vice versa.
    video_start_threshold_percentile: int
        Percentage of the start of the video in which the detections will be marked as detections. As recaps and opening credits
        only occur at the first parts of video files, this parameter can alter that threshold. So putting 20 in here means that
        if we find recurring content in the first 20% of frames of the video, it will be marked as a detection. If it's detected
        later than 20%, then the detection will be ignored.
    video_end_threshold_seconds: int
        Number of seconds threshold in which the final detection at the end of the video should end for it to count.
        Putting 15 here means that a detection at the end of a video will only be marked as a detection if the detection ends
        in the last 15 seconds of the video.
    min_detection_size_seconds: int
        Minimal amount of seconds a detection should be before counting it as a detection. As credits & recaps & previews generally
        never consist of a few seconds, it's wise to pick at least a number higher than 10.

    returns
    -------
    dictionary
        dictionary with timestamp detections in seconds list for every video file name

       {"video_filename" : [(start1, end1), (start2, end2)],
        "video_filename2" :  [(start1, end1), (start2, end2)],
        ...
       }
    """

    # if feature vector function is CNN, change resize width
    if feature_vector_function == "CNN":
        resize_width = 224

    # define the static directory names
    resized_dir_name = f"resized{resize_width}"
    feature_vectors_dir_name = (
        f"{feature_vector_function}_feature_vectors_framejump{framejump}"
    )

    # the video files used for the detection
    videos = [f for f in videos if video_functions.file_is_video(f)]

    # make sure videos are sorted, use natural sort to correctly handle case of ep1 and ep10 in file names
    videos = natsorted(videos, alg=ns.IGNORECASE)

    for file in videos:
        file = file.replace("//", "/")

    # if artifacts dir is not set, then store artifacts in dir with videos
    if artifacts_dir is None:
        artifacts_dir = "artifacts/"

    # location of the vector directory
    vectors_dir = os.path.join(
        artifacts_dir, resized_dir_name, feature_vectors_dir_name
    )

    # if there's an annotations file, get the pandas format
    if annotations is not None:
        annotations = evaluation.get_annotations(annotations)

    for file in videos:
        # set the video path files
        file_full = file
        file_resized = os.path.join(artifacts_dir, resized_dir_name, file)

        # make sure folder of experimentname exists or create otherwise
        os.makedirs(os.path.dirname(file_resized), exist_ok=True)

        # if there is no resized video yet, then resize it
        if not os.path.isfile(file_resized):
            video_functions.resize(file_full, file_resized, resize_width)

        # from the resized video, construct feature vectors
        featurevectors.construct_feature_vectors(
            file_resized, feature_vectors_dir_name, feature_vector_function, framejump
        )

    # query the feature vectors of each episode on the other episodes
    results = query_episodes_with_faiss(videos, vectors_dir)

    # evaluation variables
    total_relevant_seconds = 0
    total_detected_seconds = 0
    total_relevant_detected_seconds = 0

    all_detections = {}
    for video, result in results:
        framerate = video_functions.get_framerate(os.path.join(video))
        threshold = np.percentile(result, percentile)

        # all the detections
        below_threshold = result < threshold
        # Merge all detections that are less than 10 seconds apart
        below_threshold = fill_gaps(below_threshold, int((framerate / framejump) * 10))

        # put all the indices where values are nonzero in a list of lists
        nonzeros = [
            [i for i, value in it]
            for key, it in itertools.groupby(
                enumerate(below_threshold), key=operator.itemgetter(1)
            )
            if key != 0
        ]

        detected_beginning = []
        detected_end = []

        # loop through all the detections taking start and endpoint into account
        for nonzero in nonzeros:
            start = nonzero[0]
            end = nonzero[-1]

            # result is in first video_start_threshold% of the video
            occurs_at_beginning = end < len(result) * (
                video_start_threshold_percentile / 100
            )
            # the end of this timestamp ends in the last video_end_threshold seconds
            ends_at_the_end = end > len(result) - video_end_threshold_seconds * (
                framerate / framejump
            )

            if end - start > (
                min_detection_size_seconds * (framerate / framejump)
            ) and (  # only count detection when larger than min_detection_size_seconds seconds
                occurs_at_beginning or ends_at_the_end
            ):  # only use results that are in first part or end at last seconds

                start = start / (framerate / framejump)
                end = end / (framerate / framejump)

                if occurs_at_beginning:
                    detected_beginning.append((start, end))
                elif ends_at_the_end:
                    detected_end.append((start, end))

        detected = get_two_longest_timestamps(detected_beginning) + detected_end

        # evaluation
        if annotations is not None:
            ground_truths = evaluation.get_skippable_timestamps_by_filename(
                video, annotations
            )
            relevant_seconds, detected_seconds, relevant_detected_seconds = (
                evaluation.match_detections_precision_recall(detected, ground_truths)
            )

            total_relevant_seconds += relevant_seconds
            total_detected_seconds += detected_seconds
            total_relevant_detected_seconds += relevant_detected_seconds

        all_detections[video] = detected

    return all_detections



import cv2
import ffmpeg
import mimetypes
import os


def file_is_video(video_fn):
    """
    Checks if the given file path actually is a video file
    """
    file_type = mimetypes.guess_type(video_fn)[0]
    return file_type is not None and file_type.startswith("video")


def get_framerate(video_fn):
    """
    Return the video framerate given a video filename
    """
    video = cv2.VideoCapture(video_fn)
    return video.get(cv2.CAP_PROP_FPS)


def resize(input, output, resize_width):
    """
    Resizes a video with ffmpeg
    """
    video2 = cv2.VideoCapture(input)
    framecount = int(video2.get(cv2.CAP_PROP_FRAME_COUNT))

    if framecount > 0:
        stream = ffmpeg.input(input)
        if resize_width == 224:
            stream = ffmpeg.filter(stream, "scale", w=224, h=224)
        else:
            stream = ffmpeg.filter(stream, "scale", w=resize_width, h="trunc(ow/a/2)*2")
        stream = ffmpeg.output(stream, output)
        try:
            with open(os.devnull, "w") as devnull:
                ffmpeg.run(stream, stdout=devnull, stderr=devnull)
        except FileNotFoundError:
            raise Exception("ffmpeg not found, make sure ffmpeg is in the PATH")
    else:
        raise Exception(f"Something is wrong with the video file: {input}")



from pickle import GET
from typing import List, Callable, Any

OVERRIDE_MAP: dict[str, Callable[[Any], None]] = {}

SCAN_MOVIE = "scan_movie"
SCAN_SERIE = "scan_serie"
SCAN_GAME = "scan_game"
SCAN_MUSIC = "scan_music"
SCAN_BOOK = "scan_book"

GET_ALL_MOVIES = "get_all_movies"
GET_ALL_BOOKS = "get_all_books"
GET_ALL_PLAYLISTS = "get_all_playlists"
GET_ALL_ALBUMS = "get_all_albums"
GET_ALL_ARTISTS = "get_all_artists"
GET_ALL_TRACKS = "get_all_tracks"
GET_ALL_SERIES = "get_all_series"
GET_ALL_OTHERS = "get_all_others"
GET_ALL_CONSOLES = "get_all_console_game"
GET_ALL_GAMES = "get_all_games"

GET_ALBUM_TRACKS = "get_album_tracks"
GET_PLAYLIST_TRACKS = "get_playlist_tracks"

GET_TRACK = "get_track"
GET_ALBUM = "get_album"
GET_PLAYLIST = "get_playlist"
GET_ARTIST = "get_artist"

GET_ARTIST_ALBUMS = "get_artist_albums"
GET_ARTIST_TRACKS = "get_artist_tracks"

GET_MOVIE_DATA = "get_movie_data"
GET_SERIE_DATA = "get_serie_data"
GET_OTHER_DATA = "get_other_data"
GET_SEASON_DATA = "get_season_data"
GET_EPISODES = "get_episodes"
GET_EPISODE_DATA = "get_episode_data"
GET_TV = "get_tv"
GET_CHANNELS = "get_channels"
GET_ACTOR_DATA = "get_actor_data"


SEARCH_TV = "search_tv"
SEARCH_TRACKS = "search_tracks"
SEARCH_ALBUMS = "search_albums"
SEARCH_ARTISTS = "search_artists"
SEARCH_PLAYLISTS = "search_playlists"
SEARCH_MOVIES = "search_movies"
SEARCH_SERIES = "search_series"
SEARCH_BOOKS = "search_books"
SEARCH_OTHERS = "search_others"



def link(override_name: str):
    def decorator(handler):
        def wrapper(*args, **kwargs):
            return handler(*args, **kwargs)
        OVERRIDE_MAP[override_name] = wrapper
        return wrapper
    return decorator

def have_override(override_name):
    return override_name in OVERRIDE_MAP

def execute_override(override_name, *args, **kwargs) -> Any:
    from chocolate_app import app
    with app.app_context():
        if override_name in OVERRIDE_MAP:
            return OVERRIDE_MAP[override_name](*args, **kwargs)


from flask import Blueprint, request

from chocolate_app import DB
from chocolate_app.tables import Users
from chocolate_app.routes.api.auth import token_required
from chocolate_app.utils.utils import generate_response, Codes

profil_bp = Blueprint("profil", __name__, url_prefix="/profil")


@profil_bp.route("", methods=["GET", "POST"])
@token_required
def profil(current_user):
    if request.method == "GET":
        data = current_user.__dict__
        data.pop("_sa_instance_state")
        data.pop("password")

        return generate_response(Codes.SUCCESS, data=data)
    elif request.method == "POST":
        if not request.json:
            return generate_response(Codes.INVALID_MEDIA_TYPE)

        data = request.json

        required_keys = ["id", "name", "password", "image"]

        for key in required_keys:
            if key not in data:
                return generate_response(Codes.MISSING_DATA)

        user = Users.query.filter_by(id=data["id"]).first()

        if not user:
            return generate_response(Codes.USER_NOT_FOUND)

        user.name = data["name"]
        if len(data["password"]) > 0:
            user.password = data["password"]
        user.profile_picture = data["image"]

        DB.session.commit()

        return generate_response(Codes.SUCCESS)
    else:
        return generate_response(Codes.INVALID_METHOD)
